#!/usr/bin/env python

from bigjob import logger
import os
import uuid
import time
import traceback
import sys
from boto.ec2.connection import EC2Connection
from boto.ec2.regioninfo import RegionInfo
import boto.ec2

import saga

###############################################################################
# EC2 General
PLACEMENT_GROUP=None
SECURITY_GROUP="default"

# VM/Image specific configurations
# Recommendation Ubuntu > 1104
# apt-get install gcc python-all-dev git subversion vim
# EC2_AMI_ID="ami-c7943cae" #  ami-82fa58eb official Amazon Ubuntu 12.04 LTS (requires dev tools installation)
# EC2_AMI_ID="ami-d7f742be"
# EC2_USERNAME="ubuntu"
# EC2_KEYNAME="lsu-keypair"
# EC2_KEYNAME="MyKey"

# Authentication
# Please use ~/.boto file to configure your security credentials (if possible)
# see http://boto.readthedocs.org/en/latest/boto_config_tut.html
# 
# [Credentials]
# aws_access_key_id = <your access key>
# aws_secret_access_key = <your secret key>
#
# Alternatively you can use these two variables
AWS_ACCESS_KEY_ID=None
AWS_SECRET_ACCESS_KEY=None




class State:
    UNKNOWN="unknown"
    PENDING="pending"
    RUNNING="running"
    
    
class Service(object):
    """ Plugin for Amazon EC2 and EUCA
    
        Manages endpoint in the form of:
        
            ec2+ssh://<EC2 Endpoint>
            euca+ssh://<EUCA Endpoint>
    """

    def __init__(self, resource_url, pilot_compute_description=None):
        """Constructor"""
        self.resource_url = resource_url
        self.pilot_compute_description = pilot_compute_description
            
    
    def create_job(self, job_description):
        j = Job(job_description, self.resource_url, self.pilot_compute_description)
        return j
            
    
    def __del__(self):
        pass
    
    
    

class Job(object):
    """ Plugin for Amazon EC2 
    
        Starts VM and executes BJ agent on this VM    
        
        
        Eucalyptus on FutureGrid uses a self-signed certificate, which 1) needs to be added to boto configuration
        or 2) certificate validation needs to be disabled.
    """

    def __init__(self, job_description, resource_url, pilot_compute_description):
        
        self.job_description = job_description
        logger.debug("URL: " + str(resource_url) + " Type: " + str(type(resource_url)))
        self.resource_url = saga.Url(str(resource_url))
        self.pilot_compute_description = pilot_compute_description
        
        self.id="bigjob-" + str(uuid.uuid1())
        self.network_ip=None
        
        self.ec2_conn=None
        
        if self.resource_url.scheme == "euca+ssh" or self.resource_url.scheme == "nova+ssh":
            host = self.resource_url.host
            path = "/services/Eucalyptus"
            if self.resource_url.path!=None: 
                path = self.resource_url.path
            port = 8773
            if self.resource_url.port != None:
                port = self.resource_url.port
            region = None
            logger.debug("Host: %s, Path: %s, Port: %d"%(host, path, port))
            if self.resource_url.scheme == "euca+ssh":
                region = RegionInfo(name="eucalyptus", endpoint=host)
            elif self.resource_url.scheme == "nova+ssh":
                region = RegionInfo(name="openstack", endpoint=host)
            logger.debug("Access Key: %s Secret: %s"%(self.pilot_compute_description["access_key_id"],
                                                      self.pilot_compute_description["secret_access_key"]))
            self.ec2_conn = EC2Connection(aws_access_key_id=self.pilot_compute_description["access_key_id"],
                                          aws_secret_access_key=self.pilot_compute_description["secret_access_key"], 
                                          region=region,
                                          is_secure=False,
                                          port=port,
                                          path=path)
        else:
            aws_region = None
            if self.pilot_compute_description.has_key("region"):
                region =  self.pilot_compute_description["region"]
                logger.debug("Connect to region: %s"%(str(region)))
                aws_region = boto.ec2.get_region(region,  
                                                 aws_access_key_id=self.pilot_compute_description["access_key_id"], 
                                                 aws_secret_access_key=self.pilot_compute_description["secret_access_key"]
                                                 )

                
            self.ec2_conn = EC2Connection(aws_access_key_id=self.pilot_compute_description["access_key_id"], 
                                          aws_secret_access_key=self.pilot_compute_description["secret_access_key"],
                                          region = aws_region)
            
        self.instance = None
        
        
    def run(self):
        """ Start VM and start BJ agent via SSH on VM """
        
        """ Map fields of Pilot description to EC2 API
            { "vm_id":"ami-d7f742be",
              "vm_ssh_username":"ubuntu",
              "vm_ssh_keyname":"MyKey",
              "vm_ssh_keyfile":"<path>",
              "vm_type":"t1.micro",
              "access_key_id":"xxx",
              "secret_access_key":"xxx"
            }
        """    
            
        reservation = self.ec2_conn.run_instances(self.pilot_compute_description["vm_id"],
                                    key_name=self.pilot_compute_description["vm_ssh_keyname"],
                                    instance_type=self.pilot_compute_description["vm_type"],
                                    security_groups=[SECURITY_GROUP])
                
        self.instance = reservation.instances[0]
        self.instance_id = self.instance.id
        logger.debug("Started EC2/Eucalyptus/Nova instance: %s"%self.instance_id)
        time.sleep(5)
        self.wait_for_running()
        
        if self.resource_url.scheme != "euca+ssh" and self.resource_url.scheme != "nova+ssh":
            self.ec2_conn.create_tags([self.instance_id], {"Name": self.id})
      
        
        self.network_ip = self.instance.ip_address 
        url = "ssh://" + str(self.network_ip)
        logger.debug("Connect to: %s"%(url))

        
        # Submit job
        ctx = saga.Context("SSH")
        #ctx.type = saga.Context.SSH
        ctx.user_id = self.pilot_compute_description["vm_ssh_username"]
        ctx.user_key = self.pilot_compute_description["vm_ssh_keyfile"]

        session = saga.Session()
        session.add_context(ctx)
                
        TRIAL_MAX=30
        trials=0
        while trials < TRIAL_MAX:
            try:
                js = saga.job.Service(url, session=session)
                logger.debug("Job Description Type: " + str(type(self.job_description)))
                job = js.create_job(self.job_description)
                logger.debug("Attempt: %d, submit pilot job to: %s "%(trials,str(url)))
                job.run()
                if job.get_state()==saga.job.FAILED:
                    logger.warning("Submission failed.")
                    trials = trials + 1 
                    time.sleep(30)
                    continue
                else:
                    break
            except:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                logger.warning("Submission failed: " + str(exc_value))
                #self.__print_traceback()
                trials = trials + 1 
                time.sleep(30)
                if trials == TRIAL_MAX:
                    raise Exception("Submission of agent failed.") 
                
        logger.debug("Job State : %s" % (job.get_state())) 
        
        

    def wait_for_running(self):
        while self.get_state()!=State.RUNNING:
            time.sleep(5)
        
    
    def get_state(self):
        result = State.UNKNOWN
        try:
            self.instance.update()
            result=self.instance.state
        except:
            logger.warning("Instance not reachable/active yet...")
        return result
    
    
    def cancel(self):
        self.instance.terminate()
        
        
    ###########################################################################
    # private methods
    def __print_traceback(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        print "*** print_tb:"
        traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
        print "*** print_exception:"
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stdout)
    
 
     
if __name__ == "__main__":
    ec2_service = Service("ec2+ssh://aws.amazon.com")
    j = ec2_service.create_job("blas")
    j.run()
    print j.get_state()



""" Example application demonstrating job submission via bigjob 

    DON'T EDIT THIS FILE (UNLESS THERE IS A BUG)

    THIS FILE SHOULD NOT BE COMMITTED TO SVN WITH USER-SPECIFIC PATHS!
"""
import os
import time
import pdb
import sys
import saga

try :
    import pudb
    pudb.set_interrupt_handler()
except :
    pass

# configuration
""" This variable defines the coordination system that is used by BigJob
    e.g. 
        advert://localhost (SAGA/Advert SQLITE)
        advert://advert.cct.lsu.edu:8080 (SAGA/Advert POSTGRESQL)
        redis://localhost:6379 (Redis at localhost)
        tcp://localhost (ZMQ)
        tcp://* (ZMQ - listening to all interfaces)
"""

#COORDINATION_URL = "advert://localhost/?dbtype=sqlite3"
#COORDINATION_URL = "tcp://*"
COORDINATION_URL = "redis://10.0.1.18:6379"
#COORDINATION_URL = "redis://Oily9tourSorenavyvault@redis01.tacc.utexas.edu"
# for running BJ from local dir
sys.path.insert(0, os.getcwd() + "/../")

from bigjob import bigjob, subjob, description


def main():
    # Start BigJob

    ##########################################################################################
    # Edit parameters for BigJob
    queue="normal" # if None default queue is used
    project=None # if None default allocation is used 
    walltime=10
    processes_per_node=4
    number_of_processes = 8
    #workingdirectory=os.path.join(os.getcwd(), "agent")  # working directory for agent
    workingdirectory="agent"
    userproxy = None # userproxy (not supported yet due to context issue w/ SAGA)


    """ 
    URL of the SAGA Job Service that is used to dispatch the pilot job.
    The following URLs are accepted:

    lrms_url = "gram://oliver1.loni.org/jobmanager-pbs" # globus resource url used when globus is used. (LONI)
    lrms_url = "pbspro://louie1.loni.org" # pbspro resource url used when pbspro scheduling system is used.(Futuregrid or LSU Machines)
    lrms_url = "ssh://louie1.loni.org" # ssh resource url which launches jobs on target machine. Jobs not submitted to scheduling system.
    lrms_url = "pbs-ssh://louie1.loni.org" # Submit jobs to scheduling system of remote machine.
    lrms_url = "xt5torque://localhost" # torque resource url.

    Please ensure that the respective SAGA adaptor is installed and working
    """
    lrms_url = "ssh://boskop"
    #lrms_url = "sge://localhost"
    #lrms_url = "fork://localhost"

    ##########################################################################################

    # for i in range(99999):
    #     js = saga.job.Service (lrms_url)
    #     j  = js.run_job ("/bin/sleep 1000")
    #     print "%4d: %s" % (i, j.state)

    for i in range(99999):
        print i

        print "Start Pilot Job/BigJob at: " + lrms_url
        bj = bigjob(COORDINATION_URL)
        bj.start_pilot_job( lrms_url,
                            number_of_processes,
                            queue,
                            project,
                            workingdirectory,
                            userproxy,
                            walltime,
                            processes_per_node)

        print "Pilot Job/BigJob URL: " + bj.pilot_url + " State: " + str(bj.get_state())

        ##########################################################################################
        # Submit SubJob through BigJob
        jd = description()
        jd.executable = "/bin/echo"
        #jd.executable = "$HOME/hello.sh"
        jd.number_of_processes = "1"
        jd.arguments = ["$HELLOWORLD"]
        jd.environment = ['HELLOWORLD=hello_world']
        #jd.spmd_variation = "mpi"

        # specify an optinal working directory if sub-job should be executed outside of bigjob sandbox
        #jd.working_directory = "/tmp" 
        jd.output = "stdout.txt"
        jd.error = "stderr.txt"

        sj = subjob()
        sj.submit_job(bj.pilot_url, jd)

        #########################################
        # busy wait for completion
        while 1:
            state = str(sj.get_state())
            print "state: " + state
            if(state=="Failed" or state=="Done"):
                break
            time.sleep(2)

        ##########################################################################################
        # Cleanup - stop BigJob
        bj.cancel()
        #time.sleep(30)


""" Test Job Submission via Advert """
if __name__ == "__main__":
    main()





import os
import sys
import pilot
import traceback

""" DESCRIPTION: This example does this...
"""

# Redis password is read from the environment. The example can be run like this:
# REDIS_PASSWORD=ILikeBigJob_wITH-REdIS python examples/example_styleguide.py
# Alternatively, for tutorials, etc. REDIS_PASSWORD can be defined in /etc/profile

#------------------------------------------------------------------------------
# 
REDIS_PWD = os.environ.get('REDIS_PASSWORD')
COORD     = "redis://%s@gw68.quarry.iu.teragrid.org:6379" % REDIS_PWD
HOST      = "ssh://localhost"

#------------------------------------------------------------------------------
#
def main():
    try:
        pilot_description = pilot.PilotComputeDescription()
        pilot_description.service_url = HOST
        pilot_description.number_of_processes = 1
        pilot_description.working_directory = os.getcwd()

        pilot_service = pilot.PilotComputeService(COORD)

        ### This is broken !!! -> https://github.com/saga-project/BigJob/issues/118
        #pilotjob = pilot_service.create_pilot(pilot_compute_description)
        pilotjob = pilot_service.create_pilot(pilot_compute_description=pilot_description)

        task = pilot.ComputeUnitDescription()
        task.executable = "/bin/sleep"
        task.arguments = ["10"]

        pilotjob.submit_compute_unit(task)

        # do something useful here, wait or whatever. print some information.


        # Not sure how to cancel properly
        #   - see https://github.com/saga-project/BigJob/issues/121
        #   - see https://github.com/saga-project/BigJob/issues/131
        for i, pj in enumerate(pjs):
            print "cancel %3d" % i
            pj.cancel()

        pilot_service.cancel()

        return(0)

    except Exception, ex:
            print "AN ERROR OCCURED: %s" % ((str(ex)))
            # print a stack trace in case of an exception -
            # this can be helpful for debugging the problem
            traceback.print_exc()
            return(-1)

if __name__ == "__main__":
    sys.exit(main())



import os
import time
import pilot

try:
    import pudb
    pudb.set_interrupt_handler()
except:
    pass

#########################################################################
##
redis_password = os.environ.get('REDIS_PASSWORD')
COORD    = "redis://%s@gw68.quarry.iu.teragrid.org:6379" % redis_password
HOST     = "ssh://localhost"
##
#########################################################################

N     = 20
pjs   = []
start = time.time()
total = 0.0

for i in range(0, N):

    print "start  %3d" % i

    pilot_description = pilot.PilotComputeDescription()
    pilot_description.service_url = HOST
    pilot_description.number_of_processes = 1
    pilot_description.working_directory = os.getcwd()

    pilot_service = pilot.PilotComputeService(COORD)

    ### This is broken !!! -> https://github.com/saga-project/BigJob/issues/118
    #pilotjob = pilot_service.create_pilot(pilot_compute_description)
    pilotjob = pilot_service.create_pilot(pilot_compute_description=pilot_description)

    pjs.append(pilotjob)

    task = pilot.ComputeUnitDescription()
    task.executable = "/bin/sleep"
    task.arguments = ["10"]

    pilotjob.submit_compute_unit(task)

stop = time.time()

# see https://github.com/saga-project/BigJob/issues/121
# see https://github.com/saga-project/BigJob/issues/131

for i, pj in enumerate(pjs):
    print "cancel %3d" % i
    pj.cancel()

pilot_service.cancel()


print "time: %.1fs   rate: %.1f/s" % (stop-start, N/(stop-start))



'''
SSH-based coordination scheme between manager and agent
'''
import urlparse
import pdb
import glob
import errno
import sys
import os
import stat
import logging
import traceback
import pexpect
from pexpect import TIMEOUT
from pilot.api.api import PilotError

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
from pilot.api import State
from bigjob import logger

SSH_OPTS="-o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -o NumberOfPasswordPrompts=0"


class SSHFileAdaptor(object):
    """ BigData Coordination File Management for Pilot Store """
    
    def __init__(self, service_url, security_context=None, pilot_data_description=None):        
        self.service_url = service_url
        result = urlparse.urlparse(service_url)
        self.host = result.hostname
        self.path = result.path        
        self.user = result.username
        
        self.pilot_data_description=pilot_data_description
        
        # handle security context
        self.userkey=None
        self.security_context=security_context
         
        # try to recover key from pilot_data_description
        if self.pilot_data_description!=None and\
           self.pilot_data_description.has_key("userkey"):
            self.userkey=self.pilot_data_description["userkey"]

        logger.debug("Security Context: " + str(self.security_context))
            
        # try to recover key from security context
        if security_context!=None and security_context!="None":
            logger.debug("Attempt to restore SSH credentials from security context: " + str(security_context))
            security_context = eval(security_context)
            key=security_context["userkey"]             
            self.userkey=os.path.join(os.getcwd(), ".ssh/id_rsa")
            if os.path.exists(os.path.join(os.getcwd(),".ssh"))==False:
                os.makedirs(os.path.join(os.getcwd(), ".ssh"))
            logger.debug("Write key: " + str(type(key)) + " to: " + self.userkey)
            try:
                f = open(self.userkey, "w") 
                for i in key:
                    logger.debug("write: " + str(i))
                    f.write(i)
                f.close()
                os.chmod(self.userkey, 0600)
            except:
                self.__print_traceback()
        
        if self.__run_check()==False:
            sys.stderr.write("Unable to connect/initialize Pilot-Data. Exit BigJob.")
            raise PilotError("Unable to connect/initialize Pilot-Data. Exit BigJob.")
        
        

    def get_security_context(self):
        """ Returns security context that needs to be available on the distributed
            node in order to access this Pilot Data """
        if (self.security_context==None or self.security_context=="None") and self.pilot_data_description.has_key("userkey"):
            f = open(self.pilot_data_description["userkey"])
            key = f.readlines()
            f.close
            self.security_context = {"userkey":key}
        logger.debug("Return security context: " + str(self.security_context))        
        return self.security_context
        
        
    def initialize_pilotdata(self):
        # check whether directory exists
        try:
            command = "mkdir -p %s"%self.path 
            self.__run_ssh_command(self.userkey, self.user, self.host, command)      
        except IOError:
            self.__print_traceback()
            # directory does not exist
        self.__state=State.Running
        
        
    def get_pilotdata_size(self):
        return None
    
    
    def delete_pilotdata(self):
        self.__remove_directory(self.path)
        self.__state=State.Done
    
        
    def get_state(self):
        if self.__client.get_transport().is_active()==True:
            return self.__state
        else:
            self.__state=State.Failed
            return self.__state            
            
    def create_du(self, du_id):
        du_dir = os.path.join(self.path, str(du_id))
        logger.debug("mkdir: " + du_dir)
        command = "mkdir %s"%du_dir
        self.__run_ssh_command(self.userkey, self.user, self.host, command)
        
        
    def put_du(self, du):
        self.put_du_scp(du)
                

    def put_du_scp(self, du):
        logger.debug("Copy DU using SCP")
        du_items = du.list()
        for i in du_items.keys():     
            local_filename = du_items[i]["local"]
            remote_path = os.path.join(self.path, str(du.id), os.path.basename(local_filename))
            logger.debug("Put file: %s to %s"%(i, remote_path))                        
            if local_filename.startswith("ssh://"):
                # check if remote path is directory
                if self.__is_remote_directory(local_filename):
                    logger.warning("Path %s is a directory. Ignored."%local_filename)                
                    continue
               
                #self.__third_party_transfer(i.local_url, remote_path)                
            else:
                try:
                    if stat.S_ISDIR(os.stat(local_filename).st_mode):
                        logger.warning("Path %s is a directory. Ignored."%local_filename)                
                        continue
                except:
                    pass         
            result = urlparse.urlparse(local_filename)
            source_host = result.netloc
            source_path = result.path
            source_user = result.username
            logger.debug(str((source_host, source_path, self.host, remote_path)))
            self.__run_scp_command(self.userkey, source_user, source_host, source_path, self.user, self.host, remote_path)        
    
  
    def copy_du(self, du, pd_new):
        remote_url = pd_new.service_url + "/" + str(du.id)
        local_url =  self.service_url  + "/" + str(du.id)
        self.copy_du_to_url(du, local_url, remote_url)  
        
    
    def get_du(self, du, target_url):
        remote_url = target_url
        local_url =  self.service_url  + "/" + str(du.id)
        logger.debug("get_du(): copy %s to %s:"%(local_url, remote_url))
        self.copy_du_to_url(du, local_url, remote_url)  
        
        
    def remove_du(self, du):
        self.__remove_directory(os.path.join(self.path, du.id))
    
        
    def put_progress(self, transfered_bytes, total_bytes):
        logger.debug("Bytes transfered %d/%d"%(transfered_bytes, total_bytes))
    
        
    
    ####################################################################################
    # pure file management methods
    # used by BJ file staging
    def transfer(self, source_url, target_url):
        self.__third_party_transfer_scp(source_url, target_url)    
    
    
    def create_remote_directory(self, target_url):
        result = urlparse.urlparse(target_url)
        target_host = result.hostname
        target_path = result.path
        target_user = result.username
        logger.debug("Create directory: %s"%target_path)
        command = "mkdir %s"%target_path
        rc = self.__run_ssh_command(self.userkey, target_user, target_host, command)
        if rc==0:
            return True
        else:
            return False
                
        
    def get_path(self, target_url):
        result = urlparse.urlparse(target_url)
        return result.path
    
        
    def copy_du_to_url(self, du,  local_url, remote_url):
        self.create_remote_directory(remote_url)
        self.__third_party_transfer_scp(local_url + "/*", remote_url)
  
            
    ###########################################################################
    # Private support methods
    def __get_path_for_du(self, du):
        return os.path.join(self.path, str(du.id))
    
    
    def __remove_directory(self, path):
        """Remove remote directory that may contain files.        
        """
        if self.__exists(path):
            command = "rm -rf %s"%path
            rc = self.__run_ssh_command(self.userkey, self.user, self.host, command)
            if rc==0:
                return True
            else:
                return False
            
        
    def __is_remote_directory(self, url):
        result = urlparse.urlparse(url)
        host = result.hostname
        path = result.path
        user = result.username        
        command = "test -d %s"%path
        rc = self.__run_ssh_command(self.userkey, user, host, command)
        if rc==0:
            logger.debug("Directory found: %s"%path)
            return True
        else:
            logger.debug("Directory not found: %s"%path)
            return False
            

    def __third_party_transfer_scp(self, source_url, target_url):
        result = urlparse.urlparse(source_url)
        source_host = result.hostname
        source_path = result.path
        source_user = result.username
        if source_host==None or source_host=="":
            source_host="localhost"

        result = urlparse.urlparse(target_url)
        target_host = result.hostname
        target_path = result.path
        target_user = result.username
        if target_host==None or target_host=="":
            target_host="localhost"

        #check whether this is a local transfer
        if os.path.exists(os.path.dirname(source_path)):
            logger.debug("Target and source host are localhost. Processing: %s" %(source_path))
            expanded_path = glob.glob(source_path)
            logger.debug("Expanded path: " + str(expanded_path))
            for path in expanded_path: 
                if os.path.isdir(path):
                    logger.debug("Source path %s is directory"%path)
                    files = os.listdir(path)
                    for i in files:
                        try:
                            os.symlink(os.path.join(files, i), target_path)
                        except:
                            self.__print_traceback()
                else:
                    try:
                        os.symlink(path, os.path.join(target_path, os.path.basename(path)))
                    except:
                        self.__print_traceback()
        else:
            self.__run_scp_command(self.userkey, source_user, source_host, source_path, target_user, target_host, target_path)


 
    def __exists(self, path):
        """Return True if the remote path exists
        """
        command = "test -e %s"%path
        rc = self.__run_ssh_command(self.userkey, self.user, self.host, command)
        if rc==0:
            return True
        else:
            return False
        
 
    
    def __run_ssh_command(self, userkey, user, host, command):
        output = self.__run_ssh_command_rc(userkey, user, host, command)[0]
        return output
         
    
    def __run_ssh_command_rc(self, userkey, user, host, command):
        prefix=""
        if host != None:
            prefix = "ssh " + SSH_OPTS + " "
            if userkey != None:
                prefix = prefix + " -i " + userkey + " "
            if user!=None:
                prefix = prefix + " " + user+ "@" 
            prefix = prefix + host
        
        command = prefix + " " + command
        logger.debug(command.strip())
        child = pexpect.spawn(command.strip(), timeout=None)
        output = child.readlines()
        logger.debug("Run %s Output: %s"%(command, str(output)))
        child.close()
        return (output, child.exitstatus) 


    def __run_scp_command(self, userkey, source_user, source_host, source_path, target_user, target_host, target_path):
        logger.debug("Create scp command: source_user: %s, source_host: %s"%(source_user, source_host))
        command = "scp " + SSH_OPTS + " "
        if userkey != None:
            command = command + "-i " + userkey + " "
        
        if source_host != None and source_host!="" and source_host!="localhost":
            if source_user!=None:
                command = command + " " + source_user + "@" 
            if source_host!="localhost":
                command = command + source_host + ":"
        
        # path is a must parameter
        command = command + source_path + " "
        
        logger.debug("Create scp command: target_user: %s, target_host: %s"%(str(target_user), str(target_host)))
        if target_host != None and target_host!="" and target_host!="localhost":
            if target_user!=None:
                command = command + " " + target_user + "@" 
            if target_host!="localhost":
                command = command + target_host + ":"
            
        command = command + target_path 
        logger.debug(command)    
        child = pexpect.spawn(command.strip(), timeout=None)
        password_error=False
        try:
            child.timeout=300
            child.expect("password:",timeout=300, searchwindowsize=5024)
            password_error=True
        except Exception as ex:
            logger.debug("No password prompt error found" + str(ex))

        if password_error:
            raise PilotError("SSH key-less login not correctly setup.")
        output = child.readlines()
        logger.debug("Run %s Output: %s"%(command, str(output)))
        child.close()
        return child.exitstatus
        
    def __run_check(self):
        rc = self.__run_ssh_command_rc(self.userkey, self.user, self.host, "/bin/date")[1]
        if rc == 0:
            return True
        else:
            return False 
   
    def __print_traceback(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        print "*** print_tb:"
        traceback.print_tb(exc_traceback, limit=1, file=sys.stderr)
        print "*** print_exception:"
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stderr)
    



'''
SSH-based coordination scheme between manager and agent
'''
import urlparse
import pdb
import glob
import errno
import sys
import os
import stat
import logging
import traceback
import pexpect
import mmap

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
from pilot.api import State, PilotError
from bigjob import logger



class InMemoryFileAdaptor(object):
    """ In-memory/Local file backend for Pilot-Data 
        local:///tmp/in-memory.txt
    
    """
    
    def __init__(self, service_url, security_context=None, pilot_data_description=None):        
        self.service_url = service_url
        result = urlparse.urlparse(service_url)
        self.host = result.hostname
        self.user = result.username
        self.path = result.path   
        
        self.pilot_data_description=pilot_data_description
        
        # handle security context
        self.userkey=None
        self.security_context=security_context
         
        logger.debug("Security Context: " + str(self.security_context))
        
        self.file = open(self.path, "rw")
        self.mm = None    
                

    def get_security_context(self):
        return None
        
        
    def initialize_pilotdata(self):
        # check whether directory exists
        self.mm = mmap.mmap(self.file, 0)
        self.__state=State.Running
        
        
    def get_pilotdata_size(self):
        return None
    
    
    def delete_pilotdata(self):
        self.mm.close()
        self.__state=State.Done
    
        
    def get_state(self):
        if self.__client.get_transport().is_active()==True:
            return self.__state
        else:
            self.__state=State.Failed
            return self.__state            
            
    def create_du(self, du_id):
        self.du_id = du_id
        
        
    def put_du(self, du):
        logger.debug("Copy DU using SCP")
        du_items = du.list()
        if len(du_items)!=1:
            raise PilotError("Only DUs with 1 file supported.")
        
        local_filename = du_items[0]["local"]
        remote_path = os.path.join(self.path, str(du.id), os.path.basename(local_filename))
        logger.debug("Put file: %s to %s"%(i, remote_path))  
          
  
    def copy_du(self, du, pd_new):
        return None        
    
    def get_du(self, du, target_url):
        remote_url = target_url
        local_url =  self.service_url  + "/" + str(du.id)
        logger.debug("get_du(): copy %s to %s:"%(local_url, remote_url))
        self.copy_du_to_url(du, local_url, remote_url)  
        
        
    def remove_du(self, du):
        self.__remove_directory(os.path.join(self.path, du.id))
    
        
    def put_progress(self, transfered_bytes, total_bytes):
        logger.debug("Bytes transfered %d/%d"%(transfered_bytes, total_bytes))
    
        
    def write(self, data):
        pass
    
    
    def read(self):
        pass
    
    
    ####################################################################################
    # pure file management methods
    # used by BJ file staging
    def transfer(self, source_url, target_url):
        self.__third_party_transfer_scp(source_url, target_url)    
    
    
    def create_remote_directory(self, target_url):
        result = urlparse.urlparse(target_url)
        target_host = result.hostname
        target_path = result.path
        target_user = result.username
        logger.debug("Create directory: %s"%target_path)
        command = "mkdir %s"%target_path
        try:
            self.__run_ssh_command(self.userkey, target_user, target_host, command)
            return True
        except:
            return False
                
        
    def get_path(self, target_url):
        result = urlparse.urlparse(target_url)
        return result.path
    
        
    def copy_du_to_url(self, du,  local_url, remote_url):
        self.create_remote_directory(remote_url)
        self.__third_party_transfer_scp(local_url + "/*", remote_url)
  
            
    ###########################################################################
    # Private support methods
    def __get_path_for_du(self, du):
        return os.path.join(self.path, str(du.id))
    
    
    def __remove_directory(self, path):
        """Remove remote directory that may contain files.        
        """
        if self.__exists(path):
            command = "rm -rf %s"%path
            rc = self.__run_ssh_command(self.userkey, self.user, self.host, command)
            if rc==0:
                return True
            else:
                return False
            
        
    def __is_remote_directory(self, url):
        result = urlparse.urlparse(url)
        host = result.hostname
        path = result.path
        user = result.username        
        command = "test -d %s"%path
        rc = self.__run_ssh_command(self.userkey, user, host, command)
        if rc==0:
            logger.debug("Directory found: %s"%path)
            return True
        else:
            logger.debug("Directory not found: %s"%path)
            return False
            

    def __third_party_transfer_scp(self, source_url, target_url):
        result = urlparse.urlparse(source_url)
        source_host = result.hostname
        source_path = result.path
        source_user = result.username
        if source_host==None or source_host=="":
            source_host="localhost"

        result = urlparse.urlparse(target_url)
        target_host = result.netloc
        target_path = result.path
        target_user = result.username
        if target_host==None or target_host=="":
            target_host="localhost"

        #check whether this is a local transfer
        if os.path.exists(os.path.dirname(source_path)):
            logger.debug("Target and source host are localhost. Processing: %s" %(source_path))
            expanded_path = glob.glob(source_path)
            logger.debug("Expanded path: " + str(expanded_path))
            for path in expanded_path: 
                if os.path.isdir(path):
                    logger.debug("Source path %s is directory"%path)
                    files = os.listdir(path)
                    for i in files:
                        try:
                            os.symlink(os.path.join(files, i), target_path)
                        except:
                            self.__print_traceback()
                else:
                    try:
                        os.symlink(path, os.path.join(target_path, os.path.basename(path)))
                    except:
                        self.__print_traceback()
        else:
            self.__run_scp_command(self.userkey, source_user, source_host, source_path, target_user, target_host, target_path)


 
    def __exists(self, path):
        """Return True if the remote path exists
        """
        command = "test -e %s"%path
        rc = self.__run_ssh_command(self.userkey, self.user, self.host, command)
        if rc==0:
            return True
        else:
            return False
        
 
    
    def __run_ssh_command(self, userkey, user, host, command):
        prefix=""
        if host != None:
            prefix = "ssh " + SSH_OPTS + " "
            if userkey != None:
                prefix = prefix + " -i " + userkey + " "
            if user!=None:
                prefix = prefix + " " + user+ "@" 
            prefix = prefix + host
        
        command = prefix + " " + command
        logger.debug(command.strip())
        child = pexpect.spawn(command.strip(), timeout=None)
        output = child.readlines()
        child.close()
        return_code = child.exitstatus
        logger.debug("Run %s Output: %s Return Code: %s"%(command, str(output), str(return_code)))
        if return_code > 0:
            raise PilotError("Command: %s not successful"%command)
        return output 
    


    def __run_scp_command(self, userkey, source_user, source_host, source_path, target_user, target_host, target_path):
        logger.debug("Create scp command: source_user: %s, source_host: %s"%(source_user, source_host))
        command = "scp " + SSH_OPTS + " "
        if userkey != None:
            command = command + "-i " + userkey + " "
        
        if source_host != None and source_host!="" and source_host!="localhost":
            if source_user!=None:
                command = command + " " + source_user + "@" 
            if source_host!="localhost":
                command = command + source_host + ":"
        
        # path is a must parameter
        command = command + source_path + " "
        
        if target_host != None and target_host!="" and target_host!="localhost":
            if target_user!=None:
                command = command + " " + target_user + "@" 
            if target_host!="localhost":
                command = command + target_host + ":"
            
        command = command + target_path 
        logger.debug(command)    
        child = pexpect.spawn(command.strip(), timeout=None)
        output = child.readlines()
        logger.debug("Run %s Output: %s"%(command, str(output)))
        child.close()
        return child.exitstatus
        
    
   
    def __print_traceback(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        print "*** print_tb:"
        traceback.print_tb(exc_traceback, limit=1, file=sys.stderr)
        print "*** print_exception:"
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stderr)
    



import sys
import os
import time
import logging
import uuid
#logging.basicConfig(level=logging.DEBUG)

#sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
from pilot import PilotComputeService, PilotDataService, ComputeDataService, State
from bigjob import logger 

COORDINATION_URL = "redis://ILikeBigJob_wITH-REdIS@gw68.quarry.iu.teragrid.org:6379"

if __name__ == "__main__":      
     
    pilot_compute_service = PilotComputeService(coordination_url=COORDINATION_URL)

    # create pilot job service and initiate a pilot job
    pilot_compute_description_amazon = {
                             "service_url": 'ec2+ssh://aws.amazon.com',
                             "number_of_processes": 1,                             
                             # cloud specific attributes
                             "vm_id":"ami-d0f89fb9",
                             "vm_ssh_username":"ubuntu",
                             "vm_ssh_keyname":"luckow",
                             "vm_ssh_keyfile":"/Users/luckow/.ssh/id_rsa",
                             "vm_type":"m1.small",
                             "access_key_id":"AKIAIEXAFM6C6HRSDOFA",
                             "secret_access_key":"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"
                            }
    
    
    pilotjob = pilot_compute_service.create_pilot(pilot_compute_description_amazon)
    
    compute_data_service = ComputeDataService()
    compute_data_service.add_pilot_compute_service(pilot_compute_service)
    
    
    # start work unit
    compute_unit_description = {
            "executable": "/bin/date",
            "arguments": [],
            "number_of_processes": 1,
            "output": "stdout.txt",
            "error": "stderr.txt",   
            "input_data": [],
            "output_data": []
    }    
    
    compute_unit = compute_data_service.submit_compute_unit(compute_unit_description)
    logging.info("Finished setup of ComputeDataService. Waiting for scheduling of PD")
    compute_data_service.wait()
    
    
    logging.info("Terminate Pilot Compute/Data Service")
    compute_data_service.cancel()
    pilot_compute_service.cancel()



#!/usr/bin/env python

import textwrap
import re
import os
import pdb

from bigjob import logger
import bigjob

import saga
    
class Service(object):
    """ Plugin for SlURM    """ 

    def __init__(self, resource_url, pilot_compute_description=None):
        """Constructor"""
        self.resource_url = resource_url
        self.pilot_compute_description = pilot_compute_description
            
    
    def create_job(self, job_description):
        j = Job(job_description, self.resource_url, self.pilot_compute_description)
        return j
            
    
    def __del__(self):
        pass

class Job(object):
    """Constructor"""
    def __init__(self, job_description, resource_url, pilot_compute_description=None):
        self.job_description=job_description
        self.bootstrap_script = self.job_description.arguments[2]
        self.job_id = ""
        self.resource_url = resource_url
        self.resource_url.scheme="ssh"
        logger.debug("BigJob/SLURM: Parsing job description")        
        if pilot_compute_description == None:
            pilot_compute_description={}
        pilot_compute_description['queue'] = job_description.queue
        pilot_compute_description['project'] = job_description.project
        pilot_compute_description['working_directory'] = job_description.working_directory
        pilot_compute_description['walltime'] = job_description.wall_time_limit
        pilot_compute_description['number_of_processes'] =  job_description.total_cpu_count

        self.working_directory = pilot_compute_description["working_directory"]
        ### convert walltime in minutes to SLURM representation of time ###
        walltime_slurm="1:00:00"
        if pilot_compute_description.has_key("walltime"):    
            hrs=int(pilot_compute_description["walltime"])/60 
            minu=int(pilot_compute_description["walltime"])%60 
            walltime_slurm=""+str(hrs)+":"+str(minu)+":00"

        logger.debug("BigJob/SLURM: generate bootstrap script")
        self.bootstrap_script = textwrap.dedent("""import sys
import os
import urllib
import sys
import time
import textwrap

sbatch_file_name="bigjob_slurm_ssh"

sbatch_file = open(sbatch_file_name, "w")
sbatch_file.write("#!/bin/bash")
sbatch_file.write("\\n")
sbatch_file.write("#SBATCH -n %s")
sbatch_file.write("\\n")
sbatch_file.write("#SBATCH -J bigjob_slurm")
sbatch_file.write("\\n")
sbatch_file.write("#SBATCH -t %s")
sbatch_file.write("\\n")
sbatch_file.write("#SBATCH -A %s")
sbatch_file.write("\\n")
sbatch_file.write("#SBATCH -o %s/stdout-bigjob_agent.txt")
sbatch_file.write("\\n")
sbatch_file.write("#SBATCH -e %s/stderr-bigjob_agent.txt")
sbatch_file.write("\\n")
sbatch_file.write("#SBATCH -p %s")
sbatch_file.write("\\n")

sbatch_file.write("cd %s")
sbatch_file.write("\\n")
sbatch_file.write("python -c XX" + textwrap.dedent(\"\"%s\"\") + "XX")
sbatch_file.close()
os.system( "sbatch  " + sbatch_file_name)
""") % (str(pilot_compute_description["number_of_processes"]),str(walltime_slurm), str(pilot_compute_description["project"]), pilot_compute_description["working_directory"], pilot_compute_description["working_directory"], pilot_compute_description["queue"], pilot_compute_description["working_directory"], self.bootstrap_script)
        ### escaping characters
        self.bootstrap_script = self.bootstrap_script.replace("\"","\\\"")
        self.bootstrap_script = self.bootstrap_script.replace("\\\\","\\\\\\\\\\")
        self.bootstrap_script = self.bootstrap_script.replace("XX","\\\\\\\"")
        self.bootstrap_script = "\"" + self.bootstrap_script+ "\""
        logger.debug(self.bootstrap_script)


    def run(self):
        jd = saga.job.Description()
        jd.arguments = ["-c", self.bootstrap_script]
        jd.executable = "python"
        jd.working_directory =  self.working_directory
        jd.output = "saga_job_submission.out"
        jd.error = "saga_job_submission.err"
        # Submit job
        js = None
        js = saga.job.Service(self.resource_url)
        slurmsshjob = js.create_job(jd)
        logger.debug("Submit pilot job to: " + str(self.resource_url))
        slurmsshjob.run()
        slurmsshjob.wait()
        logger.debug("BigJob/SLURM: SSH run job finished")
        
        saga_surl = saga.Url(self.resource_url)
        sftp_url = "sftp://"
        if saga_surl.username!=None and saga_surl.username!="":
            sftp_url = sftp_url + str(saga_surl.username) + "@"
        sftp_url = sftp_url + saga_surl.host + "/"
        outfile = sftp_url + self.working_directory+'/saga_job_submission.out'        
        logger.debug("BigJob/SLURM: get outfile: " + outfile)
        out = saga.filesystem.File(outfile)
        out.copy("sftp://localhost/"+os.getcwd() + "/tmpout")
        errfile = sftp_url + self.working_directory+'/saga_job_submission.err'        
        err = saga.filesystem.File(errfile)
        err.copy("sftp://localhost/"+os.getcwd() + "/tmperr")
        
        tempfile=open(os.getcwd() + "/tmpout")
        outstr = tempfile.read().rstrip()
        tempfile.close()
        os.remove(os.getcwd() + "/tmpout")
        
        tempfile=open(os.getcwd() + "/tmperr")
        errstr = tempfile.read().rstrip()
        tempfile.close()
        os.remove(os.getcwd() + "/tmperr")

        logger.debug("Output - \n" + str(outstr))
        if ((outstr).split("\n")[-1]).split()[0] == "Submitted": 
            self.job_id=((outstr).split("\n")[-1]).split()[3]
            logger.debug("SLURM JobID: " + str(self.job_id))
        if self.job_id==None or self.job_id=="":
            raise Exception("BigJob submission via slurm+ssh:// failed: %s %s" % (outstr,errstr))


    def get_state(self):
        jd = saga.job.Description()
        jd.executable = "squeue"
        jd.arguments=["-j",self.job_id]
        jd.output="jobstate.out"   
        jd.working_directory =  self.working_directory     
        # connect to the local job service
        js = saga.job.service(self.resource_url);
        # submit the job
        job = js.create_job(jd)
        job.run()
        job.wait()
        # print the job's output
        
        outfile = 'sftp://'+saga.Url(self.resource_url).host+self.working_directory+'/jobstate.out'        
        out = saga.filesystem.File(outfile)
        out.move("sftp://localhost/"+os.getcwd() + "/tmpstate")
        
        tempfile=open(os.getcwd() + "/tmpstate")
        output = tempfile.read().rstrip()
        tempfile.close()
        os.remove(os.getcwd() + "/tmpstate")
        
        state=output.split("\n")[-1].split()[4]

        if state.upper() == "R":
            state = "Running"
        elif state.upper() == "CD" or state.upper() == "CF" or state.upper() == "CG":
            state = "Done"
        elif state.upper() == "PD":
            state = "Queue"
        else:
            state = "Unknown"
        return state 


    def cancel(self):
        logger.debug("Cancel SLURM job")
        jd = saga.job.Description()
        jd.executable = "scancel"
        jd.arguments = [self.job_id]        
        # connect to the local job service
        js = saga.job.Service(self.resource_url);
        # submit the job
        job = js.create_job(jd)
        job.run()
        # wait for the job to complete
        job.wait()




if __name__ == "__main__":
    slurm_service = Service("slurm+ssh://stampede.tacc.utexas.edu")
    j = slurm_service.create_job("")
    j.run()
    print j.get_state()



import os
import sys
import time
import logging
import uuid
import saga
import pilot
import traceback

#------------------------------------------------------------------------------
# Redis password and eThread secret key details aquired from the environment
COORD = os.environ.get('COORDINATION_URL')
ACCESS_KEY_ID = os.environ.get('ETHREAD_ACCESS_KEY_ID')
SECRET_ACCESS_KEY= os.environ.get('ETHREAD_SECRET_ACCESS_KEY')

COORD = "redis://ILikeBigJob_wITH-REdIS@gw68.quarry.iu.teragrid.org:6379"
ACCESS_KEY_ID="AKIAIEXAFM6C6HRSDOFA"
SECRET_ACCESS_KEY="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"

#------------------------------------------------------------------------------
# The coordination server
#COORD = "redis://ILikeBigJob_wITH-REdIS@gw68.quarry.iu.teragrid.org:6379"
# The host (+username) to run BigJob on
EC2url = "aws.amazon.com"
S3url = ""
# The queue on the remote system
#QUEUE = "normal"
# The working directory on the remote cluster / machine
#WORKDIR = "/home1/02554/sagatut/XSEDETutorial/%s/example1" % USER_NAME
WORKDIR = "/home/ubuntu/"
SSHKEYFILE = "/Users/luckow/.ssh/id_rsa"
#WORKDIR2 = "/home/anjani/bigjob_test/test_saga-bigjob/agent/SE_2"
# The number of jobs you want to run
NUMBER_JOBS = 1


#------------------------------------------------------------------------------
#
def main():
    try:
        # this describes the parameters and requirements for our pilot job1
        pilot_compute_description_amazon_west1 = pilot.PilotComputeDescription()
        pilot_compute_description_amazon_west1 = {
                             "service_url": 'ec2+ssh://%s' % EC2url,
                             "number_of_processes": 2,
                             "vm_id": "ami-d0f89fb9",
                             "vm_ssh_username":"ubuntu",
                             "vm_ssh_keyname":"luckow",
                             "vm_ssh_keyfile": SSHKEYFILE,
                             "vm_type":"t1.micro",
                             "region" :"us-east-1a",
                             "access_key_id":ACCESS_KEY_ID,
                             "secret_access_key":SECRET_ACCESS_KEY,
                            # "affinity_machine_label": ""
                            }

        # create a new pilot job1
        pilot_compute_service1 = pilot.PilotComputeService(COORD)
        pilotjob1 = pilot_compute_service1.create_pilot(pilot_compute_description_amazon_west1)

        # this describes the parameters and requirements for our pilot job2
        pilot_compute_description_amazon_west2 = pilot.PilotComputeDescription()
        pilot_compute_description_amazon_west2 = {
                             "service_url": 'ec2+ssh://%s' % EC2url,
                             "number_of_processes": 2,
                             "vm_id": "ami-d0f89fb9",
                             "vm_ssh_username":"ubuntu",
                             "vm_ssh_keyname":"luckow",
                             "vm_ssh_keyfile": SSHKEYFILE,
                             "vm_type":"t1.micro",
                             "region" :"us-east-1a",
                             "access_key_id":ACCESS_KEY_ID,
                             "secret_access_key":SECRET_ACCESS_KEY,
                            # "affinity_machine_label": ""
                            }

        # create a new pilot job2
        pilot_compute_service2 = pilot.PilotComputeService(COORD)
        pilotjob2 = pilot_compute_service2.create_pilot(pilot_compute_description_amazon_west2)

        # submit tasks1 to pilot job1
        tasks1 = list()
        for i in range(NUMBER_JOBS):
            task_desc1 = pilot.ComputeUnitDescription()
            task_desc1.working_directory = WORKDIR
            task_desc1.executable = '/bin/echo'
            task_desc1.arguments = ['I am task number $TASK_NO from pj1', ]
            task_desc1.environment = {'TASK_NO': i}
            task_desc1.number_of_processes = 1
            task_desc1.output = 'stdout1.txt'
            task_desc1.error = 'stderr1.txt'

            task1 = pilotjob1.submit_compute_unit(task_desc1)
            print "* Submitted task '%s' with id '%s' to %s" % (i, task1.get_id(), EC2url)
            tasks1.append(task1)

        print "Waiting for tasks to finish..."
        pilotjob1.wait()

        # submit tasks2 to pilot job2
        tasks2 = list()
        for i in range(NUMBER_JOBS):
            task_desc2 = pilot.ComputeUnitDescription()
            task_desc2.working_directory = WORKDIR
            task_desc2.executable = '/bin/echo'
            task_desc2.arguments = ['I am task number $TASK_NO from pj2', ]
            task_desc2.environment = {'TASK_NO': i}
            task_desc2.number_of_processes = 1
            task_desc2.output = 'stdout2.txt'
            task_desc2.error = 'stderr2.txt'

            task2 = pilotjob2.submit_compute_unit(task_desc2)
            print "* Submitted task '%s' with id '%s' to %s" % (i, task2.get_id(), EC2url)
            tasks2.append(task2)

        print "Waiting for tasks to finish..."
        pilotjob2.wait()
	"""
	# all compute units have finished. now we can use saga-python
        # to transfer local file to remote...
        d1 = saga.filesystem.Directory("sftp://%s/" % (HOSTNAME1))
        for task1 in tasks1:
            local_filename1 = "ex-2-stdout1-%s.txt" % (task1.get_id())
            d1.copy("%s/stdout1.txt" % (task1.get_local_working_directory()), "sftp://anjani@hotel.futuregrid.org/%s/%s" % (os.getcwd(), local_filename1))
            print "* Output for '%s' copied to: './%s'" % (task1.get_id(), local_filename1)        

	# all compute units have finished. now we can use saga-python
        # to transfer remote files to local...
        d2 = saga.filesystem.Directory("sftp://%s/" % (HOSTNAME2))
        for task2 in tasks2:
            local_filename2 = "ex-2-stdout2-%s.txt" % (task2.get_id())
            d2.copy("%s/stdout2.txt" % (task2.get_local_working_directory()), "file://localhost/%s/%s" % (os.getcwd(), local_filename2))
            print "* Output for '%s' copied to: './%s'" % (task2.get_id(), local_filename2)
	"""
        return(0)

    except Exception, ex:
            print "AN ERROR OCCURED: %s" % ((str(ex)))
            # print a stack trace in case of an exception -
            # this can be helpful for debugging the problem
            traceback.print_exc()
            return(-1)

    finally:
        # alway try to shut down pilots, otherwise jobs might end up
        # lingering in the queue
        print ("Terminating BigJob...")
        pilotjob1.cancel()
        pilot_compute_service1.cancel()
        pilotjob2.cancel()
        pilot_compute_service2.cancel()


if __name__ == "__main__":
    sys.exit(main())




# Archive Redis content and generate some primitive analytics
#
# On OS X 
# brew install zeromq freetype
# pip install pyzmq tornado pandas ipython matplotlib 

import redis
import os, sys
import pandas as pd


from bigjob import logger

# Archive the following redis urls
REDIS_URLS=["redis://ILikeBigJob_wITH-REdIS@gw68.quarry.iu.teragrid.org:6379", "redis://localhost"]


REDIS_SERVER="localhost"
REDIS_SERVER_PORT=6379
REDIS_URL_SCHEME="redis://"

class RedisDownloader(object):
    
    def __init__(self, redis_url):
        '''
        Constructor
        '''
        server_port=6379
        self.redis_url=redis_url  
        self.password=None  
        start_index =  self.redis_url.find(REDIS_URL_SCHEME)+len(REDIS_URL_SCHEME)
        server_and_port = self.redis_url[start_index:]
        password_end = server_and_port.find("@") 
        # parse out password
        if password_end != -1:
            self.password = server_and_port[:password_end]
            start_index=password_end
            server_and_port= server_and_port[(password_end+1):]
            
        # port and hostname
        if server_and_port.find(":")==-1:
            server=server_and_port
            server_port = REDIS_SERVER_PORT
        else:
            server = server_and_port.split(":")[0]
            server_port = int(server_and_port.split(":")[1])
        
        logger.debug("Connect to Redis: " + server + " Port: " + str(server_port))
        
        if self.password==None:
            self.redis_client = redis.Redis(host=server, port=server_port, db=0)
        else:
            self.redis_client = redis.Redis(host=server, port=server_port, password=self.password, db=0)

        self.pipe = self.redis_client.pipeline()
        try:
            self.redis_client.ping()
        except:
            logger.error("Please start Redis server!")
            raise Exception("Please start Redis server!")
        
        
    def get_pilots(self):
        pilots = self.redis_client.keys("bigjob:bj-*")    
        for i in pilots:
            if ":jobs:" not in i and i.count(":")==2:
                #print i
                self.pipe.hgetall(i)
        response = self.pipe.execute()
        return response;
    
    
    def get_cus(self):
        cus = self.redis_client.keys("*:jobs:*")    
        for i in cus:
            self.pipe.hgetall(i)
        response = self.pipe.execute()
        return response;
        

if __name__ == '__main__':
    if len(sys.argv)>1:
        print "Get data from " + sys.argv[1]
        rd = RedisDownloader(sys.argv[1])
        pilots = rd.get_pilots()
        cus = rd.get_cus()
        print "Loaded Redis data: %d pilots, %d cus"%(len(pilots), len(cus))
    else:
        for i in REDIS_URLS:
            rd = RedisDownloader(i)
            pilots = rd.get_pilots()
            cus = rd.get_cus()
        
        
        
        
        


import sys
import os
import time

from pilot import PilotComputeService, ComputeDataService, State


COORDINATION_URL = "redis://localhost"
NUMBER_CUS=128
NUMBER_SLOTS=16
if __name__ == "__main__":      
    
    pilot_compute_service = PilotComputeService(coordination_url=COORDINATION_URL)

    # create pilot job service and initiate a pilot job
    pilot_compute_description = {
                             "service_url": 'fork://localhost',
                             "number_of_processes": NUMBER_SLOTS,                             
                             "working_directory": os.getcwd(),
                             "number_executor_threads": 16
                            }
    
    pilotjob = pilot_compute_service.create_pilot(pilot_compute_description=pilot_compute_description)
    
    while pilotjob.get_state()!="Running":
        time.sleep(2) 
        
    start = time.time()
    # start work unit
    compute_unit_description = {
            "executable": "/bin/date",
            "arguments": [""],
            "number_of_processes": 1,            
            "output": "stdout.txt",
            "error": "stderr.txt",   
    }   
    
    cds = ComputeDataService()
    cds.add_pilot_compute_service(pilot_compute_service)
    
    unitservice = pilotjob
    
    for i in range(0,NUMBER_CUS):
        compute_unit = unitservice.submit_compute_unit(compute_unit_description)
    
    
    print("Finished setup. Waiting for scheduling of CU")
    unitservice.wait()
    
    
    runtime=time.time()-start
    print("Number Slots, Number CUs, Runtime, Throughput")
    print("%d,%d,%f,%f"%(NUMBER_SLOTS,NUMBER_CUS,runtime, runtime/NUMBER_CUS))
    
    print("Terminate Pilot Compute and Compute Data Service")
    pilot_compute_service.cancel()



# -*- coding: utf-8 -*-
# <nbformat>3.0</nbformat>

# <markdowncell>

# # Generating BigJob Usage Statistics out of Redis entries
# Read `cus` and `pilots` from Redis

# <codecell>

import pandas as pd
import matplotlib.pyplot as plt
import os, sys
import archive
import datetime
import ast

# <codecell>

# Attempt to restore old data frame
cus_df = None
pilot_df = None
if os.path.exists("cus.df") and os.path.exists("pilot.df"):
    cus_df = pd.load("cus.df") #pd.read_csv("cus.csv", index_col=0, parse_dates=False, date_parser=)
    pilot_df = pd.load("pilot.df") #pd.read_csv("pilot.csv", index_col=0, parse_dates=False, date_parser=), dat

    max_cus_date = cus_df.index.max()
    max_pilots_date = pilot_df.index.max()
    print "Restored data frames until %s"%max_cus_date

# <codecell>

# Download new data
# Redis Service to connect to:
# redis://ILikeBigJob_wITH-REdIS@gw68.quarry.iu.teragrid.org:6379
# redis://localhost
rd = archive.RedisDownloader("redis://ILikeBigJob_wITH-REdIS@gw68.quarry.iu.teragrid.org:6379")
#rd = archive.RedisDownloader("redis://localhost:6379")
pilots = rd.get_pilots()
cus = rd.get_cus()

# <markdowncell>

# ## Compute Units Executed per Day

# <codecell>

# make sure only new entries are loaded into data frame
max_cus_date = None
try:
    max_cus_date = cus_df.index.max()
except:
    pass
timestamp_index = []
cus_new = []
for i in cus:
    if max_cus_date == None or datetime.datetime.utcfromtimestamp(float(i["start_time"]))>max_cus_date:
        # print "add " + str(datetime.datetime.utcfromtimestamp(float(i["start_time"])))
        timestamp_index.append(datetime.datetime.utcfromtimestamp(float(i["start_time"])))
        cus_new.append(i)

#print cus_new    
if len(cus_new) > 0:
    cus_df_new = pd.DataFrame(cus_new, index=timestamp_index, columns=['Executable', 'NumberOfProcesses', "SPMDVariation", "start_time", "end_queue_time", "start_staging_time", "end_time"])
    try:
        cus_df = pd.concat([cus_df, cus_df_new])
    except:
        cus_df = cus_df_new

# <codecell>

cus_df_h = cus_df["Executable"].resample("D", how="count")
cus_df_h.plot(color='k', alpha=0.7)
plt.ylabel("Number of CUs Executed")
plt.xlabel("Day")
plt.savefig("number_cus_per_day.pdf", format="pdf", bbox_inches='tight', pad_inches=0.1)

# <markdowncell>

# ## Compute Unit Types
# 
# How many sequential versus parallel (MPI) CUs are executed?

# <codecell>

spmd = cus_df["SPMDVariation"].astype("object")
spmd[spmd.isnull()]="single"
spmd.value_counts().plot(kind="bar",  color='k', alpha=0.7)
plt.ylabel("Number of CUs")
plt.ylabel("CU SPMD Variation")
plt.savefig("cu_type.pdf", format="pdf", bbox_inches='tight', pad_inches=0.1)

# <codecell>

cus_df["Executable"].value_counts().plot(kind="bar",  color='k', alpha=0.7)
plt.ylabel("Number CUs")
plt.xlabel("CU Executable")
plt.savefig("cu_executable.pdf", format="pdf", bbox_inches='tight', pad_inches=0.1)

# <markdowncell>

# ## CU Runtime Distribution

# <codecell>

runtimes = cus_df.apply(lambda row: float(row["end_time"]) - float(row["end_queue_time"]), axis=1)
runtimes.hist(bins=50)
plt.ylabel("Number of Events")
plt.xlabel("CU Runtime (in sec)")
plt.savefig("cu_runtime.pdf", format="pdf", bbox_inches='tight', pad_inches=0.1)
runtimes.describe()

# <markdowncell>

# ## Pilots Executed per Day
# Extract pilot desciptions out of Redis entries

# <codecell>

print "Number of Pilots: %d Number CUs: %d Executed since: %s"%(len(pilots), len(cus), str(cus_df.index.min()))

# <codecell>

pilots = [i for i in pilots if i.has_key("start_time")]
max_pilot_date = None
try:
    max_pilot_date = max_pilot_date.index.max()
except:
    pass
timestamp_index = []
pilot_new = []
for i in pilots:
    if max_pilot_date == None or datetime.datetime.utcfromtimestamp(float(i["start_time"]))>max_pilot_date:
        timestamp_index.append(datetime.datetime.utcfromtimestamp(float(i["start_time"])))
        pilot_new.append(ast.literal_eval(i["description"]))

#print cus_new    
if len(pilot_new) > 0:
    pilot_df_new = pd.DataFrame(pilot_new, index=timestamp_index, columns=['service_url', "number_of_processes"])
    try:
        pilot_df = pd.concat([pilot_df, pilot_df_new])
    except:
        pilot_df = pilot_df_new

# <codecell>

pilot_df_h = pilot_df['service_url'].resample("D", how="count")
pilot_df_h.plot(kind="line",  color='k', alpha=0.7)
plt.ylabel("Number of Pilots")
plt.xlabel("Day")
plt.savefig("number_pilots.pdf", format="pdf", bbox_inches='tight', pad_inches=0.1)

# <markdowncell>

# ## Store Dataframes for later usage

# <codecell>

cus_df.save("cus.df")
pilot_df.save("pilot.df")

date_string = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
cus_df.to_csv("cus-"+date_string+".csv", index_label="Date")
pilot_df.to_csv("pilot-"+date_string+".csv", index_label="Date")




#!/usr/bin/env python

import os
import sys
from setuptools import setup
import subprocess

try:
    import saga
except:
    print "Installing BigJob and SAGA/Python."

if sys.version_info < (2, 6):
    sys.stderr.write("BigJob requires Python 2.6 and above. Installation unsuccessful!")
    sys.exit(1)

VERSION_FILE="VERSION"    
    

def update_version():
    if not os.path.isdir(".git"):
        print "This does not appear to be a Git repository."
        return
    try:
        p = subprocess.Popen(["git", "describe",
                              "--tags", "--always"],
                             stdout=subprocess.PIPE)
    except EnvironmentError:
        print "Unable to run git, not modifying VERSION"
        return
    stdout = p.communicate()[0]
    if p.returncode != 0:
        print "Unable to run git, not modifying VERSION"
        return
    
    ver = stdout.strip()
    fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'VERSION')
    f = open(fn, "w")
    f.write(ver)
    f.close()
    print "BigJob VERSION: '%s'" % ver


def get_version():
    try:
        fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'VERSION')
        f = open(fn)
        version = f.read().strip()
        f.close()
    except EnvironmentError:
        return "-1"
    return version    

    
update_version()
    
setup(name='BigJob',
      version=get_version(),
      description='P* Pilot-Job Implementation based on SAGA-Python',
      author='Andre Luckow, et al.',
      author_email='aluckow@cct.lsu.edu',
      url='https://github.com/saga-project/BigJob',
      classifiers = ['Development Status :: 5 - Production/Stable',                  
                    'Programming Language :: Python',
                    'Environment :: Console',                    
                    'Topic :: Utilities',
                    ],
      platforms = ('Unix', 'Linux', 'Mac OS'),
      packages=['bigjob', 'bigjob_dynamic', 'coordination', 'pilot', 'bigjob.job_plugin', 'pilot.api','pilot.api.compute', 'pilot.api.data', 'pilot.coordination', 
                'pilot.filemanagement', 'pilot.impl', 'pilot.scheduler', 'examples', 'api', 'bootstrap', 'cli'],
      include_package_data=True,
      # data files for easy_install
      data_files = [('', ['bigjob.conf', 'bigjob.conf']), 
                    ('', ['bigjob_agent.conf', 'bigjob_agent.conf']), 
                    ('', ['README.md', 'README.md']), 
                    ('', ['VERSION', 'VERSION'])],
      
      # data files for pip
      package_data = {'': ['*.conf']},

      install_requires=['uuid', 'threadpool', 'virtualenv', 'redis', 
                        'radical.utils', 'saga-python', 'google-api-python-client', 'python-hostlist',
                        'globusonline-transfer-api-client', 'boto>=2.2,<2.3', 'simplejson<2.1', 'pexpect', 'tldextract'],
      entry_points = {
        'console_scripts': [
            'test-bigjob = examples.example_local_single:main',
            'test-bigjob-dynamic = examples.example_manyjob_local:main',
            'pilot-cli = cli.pilot_cli:main'            
        ]
            }
)
